/**
 * 
 * All content copyright (c) 2003-2008 Terracotta, Inc.,
 * except as may otherwise be noted in a separate copyright notice.
 * All rights reserved.
 *
 */
package org.terracotta.sample.spider;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.terracotta.message.routing.FixedRouter;
import org.terracotta.message.routing.LoadBalancingRouter;
import org.terracotta.message.routing.RoundRobinRouter;
import org.terracotta.message.routing.Router;
import org.terracotta.workmanager.RoutableWorkItem;
import org.terracotta.workmanager.dynamic.DynamicWorkManager;
import org.terracotta.workmanager.statik.StaticWorkManager;

import au.id.jericho.lib.html.Source;

import commonj.work.Work;
import commonj.work.WorkEvent;
import commonj.work.WorkException;
import commonj.work.WorkItem;
import commonj.work.WorkListener;
import commonj.work.WorkManager;

public class SpiderMaster implements Runnable {

	private final static PageCache s_pageCache = new SimplePageCache();

	// the router
	private final Router<RoutableWorkItem<String>, String> m_router;

	// the work manager
	private final WorkManager m_workManager;

	// the work listener
	private final WorkListener m_workListener;

	private final Set<String> m_visitedPages = new HashSet<String>();
	private final String m_startURL;
	private final int m_maxDepth;
	private final boolean m_followExternalLinks;
	private int m_nrOfCompletedWork = 0;

	public SpiderMaster(final String startURL, final int maxDepth, final boolean followExternalLinks,
			final String[] routingIDs, final String routingIdForFailOverNode) {

		m_startURL = startURL;
		m_maxDepth = maxDepth;
		m_followExternalLinks = followExternalLinks;

		if (routingIDs == null) {
		    m_router = new RoundRobinRouter<RoutableWorkItem<String>, String>();
		    m_workListener = null;
		    m_workManager = new DynamicWorkManager(m_router);
		} else if (routingIDs.length == 1) {

			// -- SINGLE WORK QUEUE - NO ROUTING - NO RETRY
			m_router = new FixedRouter<RoutableWorkItem<String>, String>(routingIDs[0]);
			m_workListener = null;
			m_workManager = new StaticWorkManager(m_router, routingIDs);
		} else {
			// -- MULTIPLE WORK QUEUE - WITH ROUTING AND RETRY
			m_router = new LoadBalancingRouter<RoutableWorkItem<String>, String>();
			m_workListener = null; // RetryWorkListener(routingIdForFailOverNode);
            m_workManager = new StaticWorkManager(m_router, routingIDs);			
		}		
	}

	public static Source cachePage(final String link, final Source page) {
		return s_pageCache.setPage(link, page);
	}

	public void run() {
		WorkItem firstWorkItem;
		try {
			// schedule first work
			// IN: start URL
			// OUT: first pending work item
			firstWorkItem = scheduleWork(m_startURL);
		} catch (WorkException e) {
			System.err.println("WorkException: " + e.getMessage());
			return;
		}

		// keep track of all pending work
		final Set<WorkItem> pendingWork = new HashSet<WorkItem>();
		pendingWork.add(firstWorkItem);

		// loop while there still is pending work to wait for
		while (!pendingWork.isEmpty()) {
			
			// wait for any work that is completed
			Collection<?> completedWork;
			try {
				completedWork = m_workManager.waitForAny(pendingWork, WorkManager.INDEFINITE);
			} catch (InterruptedException e) {
				throw new RuntimeException(e); // bail out
			}

			// loop over all completed work
			for (Object o : completedWork) {
				WorkItem workItem = (WorkItem) o;

				// check work status (completed or rejected)
				switch (workItem.getStatus()) {
				case WorkEvent.WORK_COMPLETED:
					List<PageLink> linksToProcess = null;
					// if completed - grab the result
					SpiderWork work = ((SpiderWork) workItem.getResult());

					// grab the new links
					linksToProcess = work.getLinks();
					m_nrOfCompletedWork++;

					// remove work from the pending list
					pendingWork.remove(workItem);

					// process all the new links
					processLinks(linksToProcess, pendingWork);
					break;

				case WorkEvent.WORK_REJECTED:
					// work rejected - just remove the work
                    pendingWork.remove(workItem);
					break;

				default:
					// status is either WORK_ACCEPTED or WORK_STARTED - should never
					// happen
				    System.err.println("WorkItem is in unexpected state: " + workItem);
				    pendingWork.remove(workItem);
				}
			}
		}
		System.out.println("Completed successfully - processed " + m_nrOfCompletedWork + " pages.");
	}

	private void processLinks(final List<PageLink> linksToProcess, final Set<WorkItem> pendingWork) {
		System.out.println("Processing " + linksToProcess.size() + " link results.");
		for (PageLink link : linksToProcess) {

			// loop over all new links
			if (followLink(link)) {
				try {
					// schedule work for each link that is found
					WorkItem newWorkItem = scheduleWork(link.getUrl(), link.getDepth());

					// add the new work item to pending work list
					pendingWork.add(newWorkItem);
				} catch (WorkException e) {
					System.err.println("WorkException: " + e.getMessage());
					continue; // TODO should retry; workItem.getResult().run()
				}
			}
		}
	}

	private boolean followLink(PageLink link) {
		if (m_visitedPages.contains(link.getUrl())) {
			return false;
		} 
		if (link.getDepth() >= m_maxDepth) {
			return false;
		} 
		if (link.getUrl().endsWith(".pdf") || link.getUrl().endsWith(".gz") || link.getUrl().endsWith(".tgz")
				|| link.getUrl().endsWith(".zip") || link.getUrl().endsWith(".doc") || link.getUrl().endsWith(".ppt")) {
			// don't follow binary content links
			return false;
		} 
		if (!link.getUrl().startsWith("http") && !link.getUrl().startsWith("file")) {
			// we're only going to support http and file for now
			return false;
		} 

		return true;
	}

	private WorkItem scheduleWork(final String url, final int depth) throws WorkException {
		try {
		    m_visitedPages.add(url);

		    // create new work for the link
	        Work work = new SpiderWork(url, depth, m_followExternalLinks);

	        // schedule the work in the work manager
	        return m_workManager.schedule(work, m_workListener);
		} finally {
	        System.out.println("Submitted URL: " + url);		    
		}
	}

	private WorkItem scheduleWork(String url) throws WorkException {
		return scheduleWork(url, 1);
	}
}
