/**
 * 
 * All content copyright (c) 2003-2008 Terracotta, Inc.,
 * except as may otherwise be noted in a separate copyright notice.
 * All rights reserved.
 *
 */
package org.terracotta.message.channel;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

/**
 * Manages channels.
 */
public class ChannelManager<T, ID> 
{
	private Channel.Factory<T, ID> channelFactory;
	private final Map<ID, Channel<T, ID>> channels = new ConcurrentHashMap<ID, Channel<T, ID>>();	
	private final List<ID> routingIDs = new ArrayList<ID>();

	public ChannelManager(final Channel.Factory<T, ID> channelFactory) 
	{
		this.channelFactory = channelFactory;
	}

	public Channel<T, ID> getOrCreateChannelFor(final ID routingID) {
		if (routingID == null) throw new IllegalArgumentException("routing ID is null");

		synchronized (channels) {
			Channel<T, ID> channel;
			if (channels.containsKey(routingID)) {
				channel = channels.get(routingID);
			} else {
				channel = channelFactory.create(routingID);
				channels.put(routingID, channel);
                routingIDs.add(routingID);
	            channels.notifyAll();
                System.out.println("created new channel for routing ID: " + routingID);
			}
			return channel;
		}
	}

	public Channel<T, ID> removeChannelFor(final ID routingID) {
		if (routingID == null) throw new IllegalArgumentException("routing ID is null");
		
		Channel<T, ID> channel = null;
		synchronized (channels) {
			if (!channels.containsKey(routingID)) {	return null; }
			channel = channels.remove(routingID);
            routingIDs.remove(routingID);
            channels.notifyAll();
			System.out.println("removed the channel for routing ID: " + routingID);			
		}
		return channel;
	}
	
	/**
	 * Unregister all registered channels except the set passed in
	 * 
	 * @param routingIDs
	 */
	public void removeAllChannelsBut(final ID[] routingIDs) {
	    List<ID> ids = new ArrayList<ID>();
	    Set<ID> s = new HashSet<ID>();
	    
	    for (ID routingID : routingIDs) {
	        s.add(routingID);
	    }
	    
	    synchronized (channels) {
	        for (ID routingID : channels.keySet()) {
	            if (!s.contains(routingID)) { 
	                ids.add(routingID);
	            }
	        }	        
	        
	        for (ID routingID : ids) {
	            removeChannelFor(routingID);
	        }
	    }
	}

	public void put(final T item, final ID routingID) throws InterruptedException {
		Channel<T, ID> pipe = getOrCreateChannelFor(routingID);
		pipe.getSendPipe().put(item); // blocks if queue is full
	}
	
	public List<ID> getRoutingIDs() {
		synchronized (channels) {
		    return Collections.unmodifiableList(routingIDs);
		}
	}
	
	public void waitForAtLeastOneChannel() throws InterruptedException {
	    synchronized (channels) {
	        while (channels.isEmpty()) {
	            channels.wait();
	        }
	    }
	}

    // TODO - this will fault every channel in - not a good idea
    // TODO - not necessary to call getOrCreateChannel at the end if scanning...
    // TODO - couldn't this whole thing be pushed into the channel manager?
    public Channel<T, ID> getChannelWithShortestSendPipe() {
        Channel<T, ID> channel = null;
        int length = Integer.MAX_VALUE;
        synchronized (channels) {
            for (Channel<T , ID> i : channels.values()) {
                int size = i.getSendPipe().size();
                if (size < length) {
                    channel = i;
                    length = size;
                }
            }
        }

        return channel;
    }
    
    public Channel<T, ID> getChannelWithShortestReceivePipe() {
        Channel<T, ID> channel = null;
        int length = Integer.MAX_VALUE;
        synchronized (channels) {
            for (Channel<T, ID> i : channels.values()) {
                int size = i.getReceivePipe().size();
                if (size < length) {
                    channel = i;
                    length = size;
                }
            }
        }

        return channel;
    }

    public void reset() {
	    synchronized (channels) {
	        for (Channel<T, ID> channel : channels.values()) {
	            channel.clear();
	        }
	    }
	}
}
