/**
 * 
 * All content copyright (c) 2003-2008 Terracotta, Inc.,
 * except as may otherwise be noted in a separate copyright notice.
 * All rights reserved.
 *
 */
package org.terracotta.message.pipe;

import java.util.ArrayList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.TimeUnit;

public class BatchingPipe<T> implements Pipe<T> {
	private final long FLUSH_DELAY = 10000;

	private final int m_batchSize;
	private final Pipe<T[]> m_pipe;

	private final Object m_batchPutLock = new Object();
	private final Object m_batchGetLock = new Object();

	private transient List<T> m_batchPut = null;
	private transient T[] m_batchGet = null;

	private transient volatile int m_index = 0;
	private transient volatile long m_last = 0;
	private transient Timer m_timer = null;

	public static class Factory<T> implements Pipe.Factory<T> {
		private final int m_batchSize;
		private final Pipe.Factory<T[]> m_factory;

		public Factory(final int batchSize, final Pipe.Factory<T[]> factory) {
			m_batchSize = batchSize;
			m_factory = factory;
		}

		public Pipe<T> create() {
			return new BatchingPipe<T>(m_batchSize, m_factory.create());
		}
	}

	private class MyTimerTask extends TimerTask {
		@SuppressWarnings("unchecked")
        @Override
		public void run() {
			System.out.println("--------timer task---------");
			List<T> ready = null;

			long current = System.currentTimeMillis();
			synchronized (m_batchPutLock) {
				if (current - m_last < FLUSH_DELAY || m_batchPut == null) {
					return;
				}
				ready = m_batchPut;
				m_batchPut = null;
			}
			try {
				System.out.println("flushing buffer");
				m_pipe.put((T[]) ready.toArray());
			} catch (Exception ie) {
				// TODO
				ie.printStackTrace();
			}
		}
	}

	/**
	 * Private constructor - use BatchingQueue.Factory.
	 * 
	 * @param batchSize
	 * @param pipe
	 */
	private BatchingPipe(final int batchSize, final Pipe<T[]> pipe) {
		m_batchSize = batchSize;
		m_pipe = pipe;
	}

	@SuppressWarnings("unchecked")
    public T put(T item) throws InterruptedException {
		List<T> ready = null;
		long current = System.currentTimeMillis();

		synchronized (m_batchPutLock) {
			m_last = current;

			if (m_timer == null) {
				m_timer = new Timer(true);
				m_timer.schedule(new MyTimerTask(), 0, 100);
			}

			if (m_batchPut == null) {
				m_batchPut = new ArrayList<T>(m_batchSize);
			}

			m_batchPut.add(item);
			if (m_batchPut.size() >= m_batchSize) {
				ready = m_batchPut;
				m_batchPut = null;
			}
		}

		if (ready != null) {
			System.out.println("pushing buffer");
			m_pipe.put((T[]) ready.toArray());
		}

		return item;
	}

	public T poll(long timeout, TimeUnit unit) throws InterruptedException {
		synchronized (m_batchGetLock) {
			batchPoll(timeout, unit);
			return m_batchGet[m_index++];
		}
	}

	public T take() throws InterruptedException {
		synchronized (m_batchGetLock) {
	        if (m_batchGet == null || m_index >= m_batchGet.length) {
	            m_batchGet = m_pipe.take();
	            m_index = 0;
	        }       
			return m_batchGet[m_index++];
		}
	}
	
	public T peek()
    {
	    synchronized (m_batchGetLock) {
	        if (m_batchGet == null || m_index >= m_batchGet.length) {
	            T[] get = m_pipe.peek();
	            if (get == null) { return null; } 
	            m_batchGet = get;
	        }
	        return m_batchGet[m_index];
	    }
    }

    public int size() 
    {
		return m_pipe.size();
	}

	public void clear() 
	{
		m_pipe.clear();
	}

	private void batchPoll(long timeout, TimeUnit unit) throws InterruptedException 
	{
		if (m_batchGet == null || m_index >= m_batchGet.length) {
			m_batchGet = m_pipe.poll(timeout, unit);
			m_index = 0;
		}
	}
}
