/**
 * 
 * All content copyright (c) 2003-2008 Terracotta, Inc.,
 * except as may otherwise be noted in a separate copyright notice.
 * All rights reserved.
 *
 */
package org.terracotta.resourcepool;

import java.lang.reflect.Proxy;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.terracotta.jmx.events.ClusterEvents;
import org.terracotta.util.RemoteInvoker;

public class ResourcePoolManager
{
    private int maxSize;
    private int allocated;
    private transient Object myNodeId;
        
    private Map<ResourceManager, ResourceManagerState> resourceManagers = 
        new HashMap<ResourceManager, ResourceManagerState>();
    private Map<String, ResourceManager> resourceManagerNodeMap = 
        new HashMap<String, ResourceManager>();
    
    public ResourcePoolManager(int maxSize)
    {
        super();
        this.maxSize = maxSize;
        this.allocated = 0;        
        
        initMyNodeId();
    }
    
    private void initMyNodeId()
    {
        ClusterEvents.DefaultListener listener = new MyClusterEventsListener();
        ClusterEvents.registerListener(listener);
        try {
            listener.waitForRegistration();
        } catch (InterruptedException ie) {
            throw new RuntimeException("Caught unexpected exception: " + ie);
        }
        myNodeId = listener.getMyNodeId();
    }

    private class MyClusterEventsListener extends ClusterEvents.DefaultListener
    {
        
        public void thisNodeId(Object nodeId)
        {
            // TODO Auto-generated method stub
            
        }

        public void initialClusterMembers(Object[] arg0)
        {
            Set<String> set = new HashSet<String>();
            for (Object nodeId : arg0) {
                set.add(nodeId.toString());
            }
            
            synchronized (resourceManagers) {
                for (String nodeId : resourceManagerNodeMap.keySet()) {
                    if (!set.contains(nodeId)) {
                        System.out.println("Pool: Removing dead node: " + nodeId);
                        ResourceManager manager = resourceManagerNodeMap.get(nodeId);
                        removeResourceManager(nodeId, manager);
                    }
                }
            }
        }

        public void nodeConnected(Object arg0) { }

        public void nodeDisconnected(Object arg0)
        {
            System.out.println("Pool: " + arg0 + " left");
            ResourceManager manager = resourceManagerNodeMap.get(arg0);
            removeResourceManager(arg0, manager);
        }

        public void thisNodeConnected(Object arg0) { }

        public void thisNodeDisconnected(Object arg0) { }
    }
    
    public synchronized int getMaxSize()
    {
        return maxSize;
    }

    public synchronized void setMaxSize(int maxSize)
    {
        this.maxSize = maxSize;
        recomputeState();
    }
    
    public void addResourceManager(ResourceManager manager)
    {
        System.out.println("Pool: Adding manager");
        RemoteInvoker invoker = new RemoteInvoker(manager);
        ResourceManager proxy = (ResourceManager) Proxy.newProxyInstance(getClass().getClassLoader(),
                                                         new Class[] { ResourceManager.class },
                                                         invoker);
        ResourceManagerState state = new ResourceManagerState(myNodeId.toString(), manager, proxy);
        synchronized (resourceManagers) {            
            resourceManagers.put(manager, state);
            resourceManagerNodeMap.put(myNodeId.toString(), manager);
            recomputeState();
        }
        System.out.println("Pool: Manager added");
    }
    
    public void removeResourceManager(Object nodeId, ResourceManager manager)
    {
        synchronized(resourceManagers) {
            if (!resourceManagers.containsKey(manager)) { return; }
            
            ResourceManagerState state = resourceManagers.get(manager); 
            incrementAllocated(-state.getAllocatedCount());
            resourceManagers.remove(manager);
            resourceManagerNodeMap.remove(nodeId.toString());
            recomputeState();
        }
    }
    
    public void allocatedResources(ResourceManager manager, int resources)
    {
        System.out.println("Pool: local resources allocated now: " + resources);
        synchronized(resourceManagers) {
            ResourceManagerState state = resourceManagers.get(manager);
            if (state.allocatedResources(resources)) {     
                recomputeState();
            }
        }        
    }
    
    private void incrementAllocated(int resources)
    {
        allocated += resources;
        System.out.println("Pool: Total allocated is now: " + allocated);
    }

    private void recomputeState()
    {
        synchronized (resourceManagers) {
            if (resourceManagers.isEmpty()) { return; }
            
            int avg = maxSize / resourceManagers.size();
            int remainder = maxSize - (avg * resourceManagers.size());
            int available = maxSize - allocated;
            int desired;
            
            System.out.println("Pool: Computing state for " + resourceManagers.size() + " nodes");
            System.out.println("Pool: Avg size: " + avg);
            System.out.println("Pool: Available: " + available);
            for (ResourceManagerState state : resourceManagers.values()) {
                desired = avg;
                if (remainder > 0) {
                    desired++;
                    remainder--;
                }
                if (desired < state.getAllocatedCount()) {
                    state.setDesiredCount(desired);                     
                } else if (desired > state.getAllocatedCount()) {
                    int net = desired - state.getAllocatedCount();
                    if (net > available) {
                        net = available;
                    }
                    state.setDesiredCount(state.getAllocatedCount() + net);
                    available -= net;
                }
            }
        }
        System.out.println("Pool: State computed");
    }
    
    private class ResourceManagerState
    {
        private int currentCount;
        private int allocatedCount;
        private int requestedCount;
        final String nodeid;
        final ResourceManager manager;
        final ResourceManager proxy;
        
        public ResourceManagerState(String nodeid, ResourceManager manager, ResourceManager proxy)
        {
            super();
            allocatedCount = 0;
            requestedCount = 0;
            this.nodeid = nodeid;
            this.manager = manager;
            this.proxy = proxy;
        }
         
        public boolean allocatedResources(int resources)
        {
            boolean decremented = false;
    
            // figure out whether local resources were incremented
            // or decremented
            if (currentCount > resources) { 
                markResourcesAllocated(resources);
                decremented = true;
            }
            currentCount = resources;
            
            return decremented;
        }
        
        private void markResourcesAllocated(int resources)
        {
            incrementAllocated(resources - allocatedCount);
            allocatedCount = resources;
        }
        
        public int getRequestedCount() { return requestedCount; }
        public int getAllocatedCount() { return allocatedCount; }
        
        public void setDesiredCount(int desiredCount)
        {
            System.out.println(nodeid + ": desired resource count: " + desiredCount);
            if (desiredCount == requestedCount) { return; }

            // if net positive, then mark allocated immediately
            if (desiredCount > requestedCount) { 
                markResourcesAllocated(desiredCount);
            }
            requestedCount = desiredCount;
            System.out.println(nodeid + ": setting local resource count: " + desiredCount);
            proxy.setResources(desiredCount, ResourcePoolManager.this);
        }
        
        public ResourceManager getManager() { return manager; }
    }
}