/*
  All content copyright (c) 2003-2008 Terracotta, Inc.,
  except as may otherwise be noted in a separate copyright notice.
  All rights reserved.
*/
package org.terracotta.util;

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

public class ClusterClassLoader extends ClassLoader
{
    private static final String NAME = "ClusterClassLoader";
    
    private static Map<String, Class<?>> classes = new HashMap<String, Class<?>>();
    private static Map<String, ClassMetaData> bytes = new HashMap<String, ClassMetaData>();
    private static transient boolean loaded;
    
    public ClusterClassLoader()
    {
        super(ClusterClassLoader.class.getClassLoader());
        try {
            Class<?> namedClassLoader = findClass("com.tc.object.loaders.NamedClassLoader");
            Class<?> helper = findClass("com.tc.object.bytecode.hook.impl.ClassProcessorHelper");
            Method m = helper.getMethod("registerGlobalLoader", new Class[] { namedClassLoader }); 
            m.invoke(null, new Object[] { this });
            loaded = true;
        } catch (Exception e) {
            // tc is not present, so don't do anything fancy
            e.printStackTrace();
            loaded = false;
        }
    }

    @Override
    public Class<?> loadClass(String name) throws ClassNotFoundException
    {
        return findClass(name);
    }

    @Override
    protected Class<?> findClass(String name) throws ClassNotFoundException
    {        
        if (!loaded) {
            return getParent().loadClass(name);            
        }
        
        Class<?> result = null;
        synchronized (classes) {
            result = classes.get(name);
            if (result != null) { return result; }

            result = loadClassBytes(name);
            if (result == null) { return getParent().loadClass(name); }
            classes.put(name, result);
        }
        
        return result;
    }
    
    private Class<?> loadClassBytes(String name) throws ClassNotFoundException
    {
        ClassMetaData metaData;
        
        synchronized (bytes) {        
            try {
                File f = null;
                metaData = bytes.get(name);
                URL resource = ClassLoader.getSystemResource(name.replace('.',File.separatorChar)+".class");
                // if resource is non null, then the class is on the local fs (in the cp)
                if (resource != null) {
                    f = new File(resource.getFile());        
                }
                
               if (metaData != null) {
                   // if it's cached, but not on the fs, return it.
                   // if it's cached, but on the fs, check to see if it's 
                   // up to date
                   if (f == null || metaData.modified >= f.lastModified()) { 
                       return defineClass(name, metaData.bytes, 0, metaData.bytes.length, null);
                   }
                }
                
                // load from the fs
                byte[] classBytes = loadClassData(f);
                Class<?> result =  defineClass(name, classBytes, 0, classBytes.length, null);

                try {
                    result.getDeclaredField("$__tc_MANAGED");
                    // it's managed so cache it
                    bytes.put(name, new ClassMetaData(f.lastModified(), classBytes));
                } catch (NoSuchFieldException e) {
                    // really dumb design bad bad bad to throw an exception when the 
                    // method return value would have sufficed.  <sigh> interns.
                }
                return result;
            } catch (IOException e){
                return null;
            } 
        }
    }    

    private byte[] loadClassData(File f) throws IOException
    {    
        byte buff[] = new byte[(int) f.length()];
        DataInputStream dis = new DataInputStream(new FileInputStream(f));
        dis.readFully(buff);
        dis.close();
        return buff;
    }
     
    public String __tc_getClassLoaderName() { return NAME; }

    private static class ClassMetaData
    {
        public final long modified;
        public final byte[] bytes;
        
        public ClassMetaData(final long modified, final byte[] bytes)
        {
            super();
            this.modified = modified;
            this.bytes = bytes;
        }
    }
    
    public static void main(String[] args) throws Throwable 
    {
         ClassLoader loader = new ClusterClassLoader();
         Class<?> mainClass = loader.loadClass(args[0]); 
         Method main = mainClass.getMethod("main", new Class[] { args.getClass() });
         String[] nextArgs = new String[args.length - 1];
         System.arraycopy(args, 0, nextArgs, 0, nextArgs.length);
         try {
             main.invoke(null, new Object[] { nextArgs });
         } catch (InvocationTargetException e) {
             throw e.getCause();
         }
     }
} 
