/**
 * 
 * All content copyright (c) 2003-2008 Terracotta, Inc.,
 * except as may otherwise be noted in a separate copyright notice.
 * All rights reserved.
 *
 */
package org.terracotta.workmanager;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.terracotta.message.channel.Channel;
import org.terracotta.message.channel.ChannelManager;
import org.terracotta.message.pipe.Pipe;
import org.terracotta.message.pipe.PipeListener;
import org.terracotta.message.routing.Router;
import org.terracotta.message.topology.Topology;

import commonj.work.Work;
import commonj.work.WorkEvent;
import commonj.work.WorkException;
import commonj.work.WorkItem;
import commonj.work.WorkListener;
import commonj.work.WorkManager;

/**
 * Abstract base class for implementations of a WorkManager. Provides generic
 * implementations for the waitForAll and waitForAny methods.
 */
public abstract class AbstractWorkManager implements WorkManager 
{
    protected final Router<RoutableWorkItem<String>, String> m_router;
    protected final ChannelManager<RoutableWorkItem<String>, String> m_channelManager;
    private final RoutableWorkItem.Factory<String> m_workItemFactory = new RoutableWorkItem.Factory<String>();
    protected Set<RoutableWorkItem<String>> m_completedWork = new HashSet<RoutableWorkItem<String>>();
    protected Map<String, Set<RoutableWorkItem<String>>> m_pendingWork = new HashMap<String, Set<RoutableWorkItem<String>>>();
    protected Map<String, CompletedWorkListener> m_completedWorkListeners = new HashMap<String, CompletedWorkListener>();
    
    public AbstractWorkManager(String topology, Router<RoutableWorkItem<String>, String> router)
    {
        this(topology, router, null);
    }
    
    public AbstractWorkManager(String topology, Router<RoutableWorkItem<String>, String> router, final String[] routingIDs)
    {
        Topology<RoutableWorkItem<String>, String> t = DefaultTopologyManager.getInstance().getOrCreateTopology(topology);
        m_router = router;
        m_channelManager = t.getChannelManager();
       
        if (routingIDs != null) {
            for (String routingID : routingIDs) {              
                register(routingID);
            }
        }
    }
    
    public WorkItem schedule(final Work work) {
        return schedule(work, null);
    }

    public WorkItem schedule(final Work work, final WorkListener listener) 
    {
        RoutableWorkItem<String> workItem = null;
        try {
            m_channelManager.waitForAtLeastOneChannel();
            Channel<RoutableWorkItem<String>, String> channel = m_router.getRoute(m_channelManager);
            String routingID = channel.getRoutingID();
            workItem = m_workItemFactory.create(work, listener, null);
            addWorkItemToPending(workItem, routingID);
            channel.getSendPipe().put(workItem);
        } catch (InterruptedException e) {
            interrupted(workItem);
        }
        return workItem;
    }
    
    private WorkItem schedule(RoutableWorkItem<String> workItem)
    {
        try {
            m_channelManager.waitForAtLeastOneChannel();
            Channel<RoutableWorkItem<String>, String> channel = m_router.getRoute(m_channelManager);
            String routingID = channel.getRoutingID();
            addWorkItemToPending(workItem, routingID);
            channel.getSendPipe().put(workItem);                    
        } catch (InterruptedException ie) {
            interrupted(workItem);
        }
        return workItem;
    }
    
    private void addWorkItemToPending(RoutableWorkItem<String> workItem, String routingID)
    {
        Set<RoutableWorkItem<String>> pendingSet = getPendingSet(routingID);
        synchronized (pendingSet) {
            pendingSet.add(workItem);
        }        
    }

    private void interrupted(RoutableWorkItem<String> workItem) 
    {
        workItem.setStatus(WorkEvent.WORK_REJECTED, new WorkException("thread interrupted"));
    }

    public synchronized void register(String routingID)
    {
        Channel<RoutableWorkItem<String>, String> channel = m_channelManager.getOrCreateChannelFor(routingID);
        CompletedWorkListener listener = new CompletedWorkListener(routingID, channel.getReceivePipe()); 
        m_completedWorkListeners.put(routingID, listener);
        listener.start();
        notifyAll();
        
        System.out.println("registered channel with ID: " + routingID);
    }
    
    public void unregisterAllBut(String[] routingIDs)
    {
        m_channelManager.removeAllChannelsBut(routingIDs);
    }

    public synchronized void unregister(String routingID)
    {
        m_channelManager.removeChannelFor(routingID);
        CompletedWorkListener listener = m_completedWorkListeners.remove(routingID);
        listener.shutdown();
        
        System.out.println("unregistered channel with ID: " + routingID);
    }

    public Set<RoutableWorkItem<String>> getAllCompletedWork() 
    {
        return m_completedWork;
    }

    public void reset() 
    {
        m_channelManager.reset();
        m_completedWork.clear();
    }

    private void reroute(String routingID)
    {
        Set<RoutableWorkItem<String>> pendingSet = getPendingSet(routingID);
        synchronized (pendingSet) {
            for (RoutableWorkItem<String> workItem : pendingSet) {
                System.out.println("rerouting work item: " + workItem.getResult());
                schedule(workItem);
                System.out.println("rerouted work item: " + workItem.getResult());
            }
        }
    }
    
    @SuppressWarnings("unchecked")
    public boolean waitForAll(Collection workItems, long timeout) throws InterruptedException {
        final int nrOfPendingWorkItems = workItems.size();
        int nrOfCompletedWorkItems = 0;
        while (true) {
            WorkItem workItem = waitForAny(timeout);
            if (workItem == null) {
                return false;
            }
            nrOfCompletedWorkItems++;
            if (nrOfPendingWorkItems == nrOfCompletedWorkItems) {
                break;
            }
        }
        return true;
    }

    @SuppressWarnings("unchecked")
    public Collection<WorkItem> waitForAny(final Collection workItems, final long timeout) throws InterruptedException {
        final Collection<WorkItem> completedWorkItems = new ArrayList<WorkItem>();

        // TODO: cheating now by only wrapping a single completed item in a
        // collection
        WorkItem workItem = waitForAny(timeout);
        if (workItem == null) {
            return Collections.emptyList();
        }
        completedWorkItems.add(workItem);
        return completedWorkItems;
    }

    public WorkItem waitForAny(long timeout) throws InterruptedException {
        synchronized (m_completedWork) {
            while (true) {
                for (WorkItem workItem : m_completedWork) {
                    if (workItem.getStatus() == WorkEvent.WORK_COMPLETED || 
                            workItem.getStatus() == WorkEvent.WORK_REJECTED) {
                        m_completedWork.remove(workItem);
                        return workItem;
                    }
                }
                if (timeout == WorkManager.IMMEDIATE) {
                    return null;
                }
                if (timeout == WorkManager.INDEFINITE) {
                    int startSize = m_completedWork.size();
                    while (m_completedWork.size() == startSize) { 
                        try {
                            m_completedWork.wait();
                        } catch (InterruptedException ie) {
                            ie.printStackTrace();
                            Thread.currentThread().interrupt();
                            throw ie;
                        }
                    }
                }
            }
        }
    }
    
    private Set<RoutableWorkItem<String>> getPendingSet(String routingID)
    {
        Set<RoutableWorkItem<String>> pendingSet = null;
        synchronized (m_pendingWork) {
            pendingSet = m_pendingWork.get(routingID);
            if (pendingSet == null) {
                pendingSet = new HashSet<RoutableWorkItem<String>>();
                m_pendingWork.put(routingID, pendingSet);
            }
            return pendingSet;
        }
    }
    
    private class CompletedWorkListener extends PipeListener<RoutableWorkItem<String>, String> {

        public CompletedWorkListener(final String routingID, final Pipe<RoutableWorkItem<String>> pipe) {
            super(routingID, pipe);
        }

        @Override
        public void stopped()
        {
            reroute(routingID);
        }

        @Override
        public boolean event(RoutableWorkItem<String> workItem) throws Exception {
            
            /*
             * If t comes back with status of ACCEPTED, it ran on the fast path, so it's
             * real status is COMPLETED, so we set that now (not on the remote node)
             */
            if (workItem.getStatus() == WorkEvent.WORK_ACCEPTED) {
                workItem.setStatus(WorkEvent.WORK_COMPLETED, null);
            }

            /*
             * Now call the listener
             */
            workItem.fireListener(); 

            /*
             * The item is completed, flag it as done
             */
            Set<RoutableWorkItem<String>> pendingSet = getPendingSet(routingID);
            synchronized (pendingSet) {
                pendingSet.remove(workItem);
            }
            synchronized (m_completedWork) {
                m_completedWork.add(workItem);
                m_completedWork.notifyAll();
            }
            return true;
        }
    }
}
